import ast
from dolang.grammar import str_expression as to_source
from dolang.symbolic import stringify, list_variables, list_symbols
from dolang.grammar import Tree


def test_parsing():
    from dolang.grammar import str_expression, sanitize
    from dolang.symbolic import parse_string
    e = parse_string("s + a(0) + b[t-1] + b[t] + b[t+1]")
    print(e.pretty())

    e = parse_string("chi*n^eta*c^sigma - w(1)")
    print(e.pretty())

    e = parse_string("chi*n^eta*c^sigma - w(1) | 0.01 <= n <= 1.0")
    print(e.pretty())

    e = parse_string("chi*n^eta*c^sigma - w(1) ⟂ 0.01 <= n <= 1.0")
    print(e.pretty())

    e = parse_string("i = exp(z)*k^alpha*n^(1-alpha) - (m)^(-1/sigma)")
    print(e.pretty())
    f = sanitize(e, variables=['m'])
    print(str_expression(f))

    s = "i = exp(z)*k^alpha*n^(1-alpha) - (m)^(-1/sigma)"
    e = parse_string(s)

    s = "i = exp(z)*k^alpha*n^(1-alpha) - (m)^(-1/sigma)"
    # vars = ['z', 'p', 'k', 'n', 'i', 'm', 'V', 'u', 'y', 'c', 'rk', 'w', 'y', 'c']
    # print("HI")
    # v = sanitize(s, variables=vars)
    # print(v)
    e = parse_string(s)
    print( str_expression(e) )

def test_parse_string():
    from dolang.symbolic import parse_string
    e = parse_string('sin(a(1)+b+f(1)+f(4)+a[t+1])')
    assert isinstance(e, Tree)
    s = to_source(e)
    assert (s == "sin(a[t+1] + b + f[t+1] + f[t+4] + a[t+1])")


# def test_list_symbols_debug():
#     from dolang.symbolic import parse_string
#     e = parse_string('sin(a(1)+b+f(1)+f(4)+a(1)+a+a[t+1]+cos(0)')
#     l = ListSymbols(known_functions=['sin', 'f'])
#     l.visit(e)
#     # note that cos is recognized as variable
#     assert (l.variables == [(('a', 1), 4), (('a', 1), 22), (('cos', 0), 37)])
#     assert (l.constants == [('b', 9), ('a', 27)])
#     assert (l.functions == [('sin', 0), ('f', 11), ('f', 16)])
#     assert (l.problems == [['a', 0, 29, 'incorrect subscript']])


def test_list_symbols():
    from dolang.symbolic import parse_string
    e = parse_string('sin(a(1)+b+f(1)+f(4)+a(1))+cos(0)')
    ll = list_symbols(e)
    print(ll.variables)
    print(ll.parameters)
    # cos is recognized as a usual function
    assert (ll.variables == [('a', 1), ('f', 1), ('f', 4)])
    assert (ll.parameters == ['b'])




def test_list_variables():
    from dolang.symbolic import parse_string
    e = parse_string('sin(a(1)+b+f(1)+f(4)+sin(a)+k*cos(a(0)))')
    list_variables(e)
    assert (list_variables(e) == [('a', 1), ('f', 1), ('f', 4), ('a', 0)])


def test_sanitize():

    from dolang.symbolic import sanitize, parse_string

    s = 'sin(a(1)+b+a+f(-1)+f(4)+a(1))'

    expected = "sin(a[t+1] + b + a[t] + f[t-1] + f[t+4] + a[t+1])"
    assert (sanitize(s, variables=['a', 'f']) == expected)

    # # we also deal with = signs, and convert to python exponents
    # assert (sanitize("a(1) = a^3 + b") == "a(1) == (a) ** (3) + b")


def test_stringify():

    from dolang.symbolic import parse_string
    s = 'sin(a(1) + b + a(0) + f(-1) + f(4) + a(1))'
    enes = stringify(s)
    assert ( enes == "sin(a__1_ + b_ + a__0_ + f_m1_ + f__4_ + a__1_)")


def test_time_shift():

    from dolang.symbolic import time_shift, stringify_parameter
    e = 'sin(a(1) + b + a(0) + f(-1) + f(4) + a(1))'

    enes = stringify(time_shift(e, +1))
    assert (
        (enes) == "sin(a__2_ + b_ + a__1_ + f__0_ + f__5_ + a__2_)")



import re
import ast

def eval_scalar(tree):
    try:
        if isinstance(tree, ast.Num):
            return tree.n
        elif isinstance(tree, ast.UnaryOp):
            if isinstance(tree.op, ast.USub):
                return -tree.operand.n
            elif isinstance(tree.op, ast.UAdd):
                return tree.operand.n
        else:
            raise Exception("Don't know how to do that.")
    except:
        raise Exception("Don't know how to do that.")


def parse(s): return ast.parse(s).body[0].value

class ExprVisitor(ast.NodeVisitor):

    def __init__(self, variables):
        self.variables = variables

    def visit_Call(self, call):
        name = call.func.id
        if name in self.variables:
            assert(len(call.args) == 1)
            n = eval_scalar(call.args[0])
            return self.visit_Variable((name, n))
        else:
            return self.visit_RCall(call)

    def visit_RCall(self, call):
        return self.generic_visit(call)

    def visit_Name(self, cname):
        name = cname.id
        if name in self.variables:
            return self.visit_Variable((name, 0))
        else:
            return self.visit_RName(cname)

    def visit_RName(self, name):
        return self.generic_visit(name)

class ExprTransformer(ast.NodeTransformer):

    def __init__(self, variables):
        self.variables = variables

    def visit_Call(self, call):
        name = call.func.id
        if name in self.variables:
            assert(len(call.args) == 1)
            n = eval_scalar(call.args[0])
            return self.visit_Variable((name, n))
        else:
            return self.generic_visit(call)

    def visit_Name(self, cname):
        name = cname.id
        if name in self.variables:
            return self.visit_Variable((name, 0))
        else:
            return self.generic_visit(cname)


class TimeShift(ExprVisitor):

    def __init__(self, variables, shift):
        self.variables = variables
        self.shift = shift

    def visit_Variable(self, tvar):
        name, t = tvar
        return parse( "{}({})".format(name,t+self.shift))


class Apply(ExprVisitor):

    def __init__(self, variables, fun):
        self.variables = variables
        self.fun = fun

    def visit_Variable(self, tvar):
        return self.fun(tvar)


reg_rad = re.compile("([^_]+)")
reg_sep = re.compile("(&|_)")
reg_bar = re.compile("(.*)(bar|star)")

gl = ['alpha', 'beta', 'gamma', 'delta', 'eta','epsilon', 'iota', 'kappa',
'lambda', 'mu', 'nu', 'rho','pi', 'sigma', 'tau','theta','upsilon','omega','phi','psi','zeta', 'xi', 'chi',
'Gamma', 'Delta', 'Lambda', 'Sigma','Theta','Upsilon','Omega','Xi' , 'Pi' ,'Phi','Psi' ]
gl_special = {
    'lam': '\\lambda'
}
greek_letters = dict([ (x,'\\' + x ) for x in gl ])
greek_letters.update(gl_special)

def greekify(expr):
    m = reg_bar.match(expr)
    if m:
        expr = m.group(1)
        suffix = m.group(2)
    else:
        suffix = None
    if expr in greek_letters:
        res = greek_letters[expr]
    else:
        res = expr
    if suffix=='bar':
        res = "\\overline{{{}}}".format(res)
    elif suffix == 'star':
        res = "{}^{{\\star}}".format(res)
    return res
# greekify('zbar')


def split_name_into_parts(a):
    s = a.replace('__','&')
    m = reg_rad.findall(a)
    rad = m[0]
    cont = m[1:]
    m = reg_sep.findall(s)
    exponents = []
    indices = []
    for i in range(len(cont)):
        if m[i] == '_':
          indices.append(cont[i])
        else:
          exponents.append(cont[i])
    return [rad, indices, exponents]

def name_to_latex(name, date=None):
    [rad, indices, exponents] = split_name_into_parts(name)
    rad = greekify(rad)
    indices = [greekify(r) for r in indices]
    exponents = [greekify(r) for r in exponents]
    up =  '{' + str.join(',', exponents) + '}'
    down =  '{' + str.join(',', indices) + '}'
    sup = '^{0}'.format(up) if up != '{}'  else ''
    sdown = '^{0}'.format(down) if down != '{}' else ''

    up =  '{' + str.join(',', exponents) + '}'

    if date is not None and date != 'S':
        if date == 0:
            times = 't'
        elif date >0:
            times = 't+' + str(date)
        elif date <0:
            times = 't-' + str(-date)
        indices = indices + [times]
    # else:
        # raise(Exception('Time variable {0} has unknown date : {1}.'.format(name,date)))

    down =  '{' + str.join(',', indices) + '}'

    if len(up)>2 and len(down)>2:
        resp = '{0}_{1}^{2}'.format(rad,down,up)
    elif len(up)>2:
        resp = '{0}^{1}'.format(rad,up)
    elif len(down)>2:
        resp = '{0}_{1}'.format(rad,down)
    else:
        resp = '{0}'.format(rad)

    if date == 'S':
        resp = '\\overline{' + resp + '}'
    return resp



class LatexVisitor(ExprVisitor):

    def prec(self, n):
        return getattr(self, 'prec_'+n.__class__.__name__, getattr(self, 'generic_prec'))(n)

    def visit_Variable(self, tvar):
        return name_to_latex(tvar[0], tvar[1])

    def visit_RCall(self, n):
        func = self.visit(n.func)
        args = ', '.join(map(self.visit, n.args))
        if func == 'sqrt':
            return '\\sqrt{%s}' % args
        else:
            return r'\operatorname{%s}\left(%s\right)' % (func, args)

    def prec_Call(self, n):
        return 1000
    def prec_RCall(self, n):
        return 1000
    def visit_RName(self, n):
        return name_to_latex( n.id )
    def prec_RName(self, n):
        return 1000

    def prec_Name(self, n):
        return 1000

    def visit_UnaryOp(self, n):
        if self.prec(n.op) > self.prec(n.operand):
            return r'%s \left(%s\right)' % (self.visit(n.op), self.visit(n.operand))
        else:
            return r'%s %s' % (self.visit(n.op), self.visit(n.operand))

    def prec_UnaryOp(self, n):
        return self.prec(n.op)

    def visit_BinOp(self, n):
        if self.prec(n.op) > self.prec(n.left):
            left = r'\left(%s\right)' % self.visit(n.left)
        else:
            left = self.visit(n.left)
        if self.prec(n.op) > self.prec(n.right):
            right = r'\left(%s\right)' % self.visit(n.right)
        elif self.prec(n.op) == self.prec(n.right) and isinstance(n.op, ast.Sub):
            right = r'\left(%s\right)' % self.visit(n.right)
        else:
            right = self.visit(n.right)
        if isinstance(n.op, ast.Div):
            return r'\frac{%s}{%s}' % (self.visit(n.left), self.visit(n.right))
        elif isinstance(n.op, ast.FloorDiv):
            return r'\left\lfloor\frac{%s}{%s}\right\rfloor' % (self.visit(n.left), self.visit(n.right))
        elif isinstance(n.op, ast.Pow):
            return r'%s^{%s}' % (left, self.visit(n.right))
        else:
            return r'%s %s %s' % (left, self.visit(n.op), right)

    def prec_BinOp(self, n):
        return self.prec(n.op)

    def visit_Sub(self, n):
        return '-'

    def prec_Sub(self, n):
        return 300

    def visit_Add(self, n):
        return '+'

    def prec_Add(self, n):
        return 300

    def visit_Mult(self, n):
        return '\\;'

    def prec_Mult(self, n):
        return 400

    def visit_Mod(self, n):
        return '\\bmod'

    def prec_Mod(self, n):
        return 500

    def prec_Pow(self, n):
        return 700

    def prec_Div(self, n):
        return 400

    def prec_FloorDiv(self, n):
        return 400

    def visit_LShift(self, n):
        return '\\operatorname{shiftLeft}'

    def visit_RShift(self, n):
        return '\\operatorname{shiftRight}'

    def visit_BitOr(self, n):
        return '\\operatorname{or}'

    def visit_BitXor(self, n):
        return '\\operatorname{xor}'

    def visit_BitAnd(self, n):
        return '\\operatorname{and}'

    def visit_Invert(self, n):
        return '\\operatorname{invert}'

    def prec_Invert(self, n):
        return 800

    def visit_Not(self, n):
        return '\\neg'

    def prec_Not(self, n):
        return 800

    def visit_UAdd(self, n):
        return '+'

    def prec_UAdd(self, n):
        return 800

    def visit_USub(self, n):
        return '-'

    def prec_USub(self, n):
        return 800

    def visit_Num(self, n):
        return str(n.n)

    def prec_Num(self, n):
        return 1000

    def generic_visit(self, n):
        if isinstance(n, ast.AST):
            return r'' % (n.__class__.__name__, ', '.join(map(self.visit, [getattr(n, f) for f in n._fields])))
        else:
            return str(n)

    def generic_prec(self, n):
        return 0

def expr2tex(variables, s):
    pt = ast.parse(s).body[0].value
    return LatexVisitor(variables).visit(pt)

def eq2tex(variables, s):

    expr = s.replace('==', '=').replace('=','==')
    if '==' in expr:
        lhs, rhs = [expr2tex(variables,str.strip(e)) for e in str.split(expr,'==')]
        return "{} = {}".format(lhs, rhs)
    else: return expr2tex(variables,expr)

# ast.dump(ast.parse("a == b"))



# all symbolic functions take ast expression trees (not expressions) as input
# this one can be constructed as : ast.parse(s).body[0].value

import ast

from ast import UnaryOp, UAdd, USub, Name, Load, Call
from ast import NodeTransformer


class Compare:

    def __init__(self):
        self.d = {}

    def compare(self, A, B):
        if isinstance(A, ast.Name) and (A.id[0] == '_'):
            if A.id not in self.d:
                self.d[A.id] = B
                return True
            else:
                return self.compare(self.d[A.id], B)
        if not (A.__class__ == B.__class__): return False
        if isinstance(A, ast.Name):
            return A.id == B.id
        elif isinstance(A, ast.Call):
            if not self.compare(A.func, B.func): return False
            if not len(A.args)==len(B.args): return False
            for i in range(len(A.args)):
                if not self.compare(A.args[i], B.args[i]): return False
            return True
        elif isinstance(A, ast.Num):
            return A.n == B.n
        elif isinstance(A, ast.Expr):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Module):
            if not len(A.body)==len(B.body): return False
            for i in range(len(A.body)):
                if not self.compare(A.body[i], B.body[i]): return False
            return True
        elif isinstance(A, ast.BinOp):
            if not isinstance(A.op, B.op.__class__): return False
            if not self.compare(A.left, B.left): return False
            if not self.compare(A.right, B.right): return False
            return True
        elif isinstance(A, ast.UnaryOp):
            if not isinstance(A.op, B.op.__class__): return False
            return self.compare(A.operand, B.operand)
        elif isinstance(A, ast.Subscript):
            if not self.compare(A.value, B.value): return False
            return self.compare(A.slice, B.slice)
        elif isinstance(A, ast.Index):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Compare):
            if not self.compare(A.left, B.left): return False
            if not len(A.ops)==len(B.ops): return False
            for i in range(len(A.ops)):
                if not self.compare(A.ops[i], B.ops[i]): return False
            if not len(A.comparators)==len(B.comparators): return False
            for i in range(len(A.comparators)):
                if not self.compare(A.comparators[i], B.comparators[i]): return False
            return True
        elif isinstance(A, ast.In):
            return True
        elif isinstance(A, (ast.Eq, ast.LtE)):
            return True
        else:
            print(A.__class__)
            raise Exception("Not implemented")


def compare(a,b):
    comp = Compare()
    val = comp.compare(a,b)
    d = comp.d
    return val

def match(m,s):
    comp = Compare()
    val = comp.compare(m,s)
    d = comp.d
    if len(d) == 0:
        return val
    else:
        return d

known_functions = ['log','exp','sin','cos']

import ast

class ListNames(ast.NodeVisitor):
    def __init__(self):
        self.found = []
    def visit_Name(self, name):
        self.found.append(name.id)

def get_names(expr):
    ln = ListNames()
    ln.visit(expr)
    return [e for e in ln.found]

def eval_scalar(tree):
    try:
        if isinstance(tree, ast.Num):
            return tree.n
        elif isinstance(tree, ast.UnaryOp):
            if isinstance(tree.op, ast.USub):
                return -tree.operand.n
        else:
            raise Exception("Don't know how to do that.")
    except:
        raise Exception("Don't know how to do that.")


class ExpressionChecker(ast.NodeVisitor):

    def __init__(self, spec_variables, known_functions, known_constants):
        self.spec_variables = spec_variables
        self.known_functions = known_functions
        self.known_constants = known_constants
        self.functions = []
        self.variables = []
        self.problems = []

    def visit_Call(self, call):
        name = call.func.id
        colno = call.func.col_offset
        if name in self.spec_variables:
            try:
                assert(len(call.args)==1)
                n = eval_scalar(call.args[0])
                allowed_timing = self.spec_variables[name]
                if allowed_timing is None or (n in allowed_timing):
                    self.variables.append((name, n, call.func.col_offset))
                else:
                    self.problems.append([name,n,colno,'incorrect_timing',allowed_timing])
            except Exception as e:
                print(e)
                self.problems.append([name,None,colno,'timing_error'])

        elif name in self.known_functions:
            self.functions.append((name, colno))
            for e in call.args:
                self.visit(e)
        else:
            self.problems.append([name, None, colno,'unknown_function'])

    def visit_Name(self, name):
        # colno = name.colno
        colno = name.col_offset
        n = 0
        name = name.id
        if name in self.spec_variables:
            allowed_timing = self.spec_variables[name]
            if (allowed_timing is None) or (n in allowed_timing):
                self.variables.append((name, n, colno))
            else:
                self.problems.append([name,n,colno,'incorrect_timing',allowed_timing])
        elif name not in self.known_constants:
            self.problems.append([name,0,colno,'unknown_variable'])

def check_expression(expr, spec_variables, known_functions=[]):

    from dolo.compiler.language import functions, constants
    func = list(functions.keys()) + known_functions

    ch = ExpressionChecker(spec_variables, func, constants)
    ch.visit(expr)
    return dict(
        functions=ch.functions,
        variables=ch.variables,
        problems=ch.problems
    )



import math

functions = {
    'log': math.log,
    'exp': math.exp,
    'sin': math.sin,
    'cos': math.cos,
    'tan': math.tan,
    'atan': math.atan,
    'tanh': math.tanh,
    'atanh': math.atanh,
    'min': min,
    'max': max,
    'minimum': min,
    'maximum': max,
    'abs': abs
}

constants = {
    'pi': math.pi,
}



import ast
from ast import Assign, arg, FunctionDef, Module, Store, Subscript, Name, Load, Index, Num
from ast import arguments as ast_arguments

from dolang.grammar import tree_to_ast
from dolang.factory import FlatFunctionFactory
from dolang.symbolic import parse_string
from dolang.codegen import to_source

import sys

def compile_factory(fff: FlatFunctionFactory):
    arguments = [*fff.arguments.keys()]
    funname = fff.funname

    unpacking = []
    for i, (arg_group_name, arg_group) in enumerate(fff.arguments.items()):
        for pos, sym in enumerate(arg_group):
            rhs = Subscript(value=Name(id=arg_group_name, ctx=Load()),
                            slice=Index(Num(pos)), ctx=Load())
            val = Assign(targets=[Name(id=sym, ctx=Store())], value=rhs)
            unpacking.append(val)

    body = []

    for (k, neq) in fff.preamble.items():
        tree = parse_string(neq)
        val = tree_to_ast(tree).value
        line = Assign(targets=[Name(id=k, ctx=Store())], value=val)
        body.append(line)

    for n, (k, neq) in enumerate(fff.content.items()):
        tree = parse_string(neq)
        val = tree_to_ast(tree).value
        line = Assign(targets=[Name(id=k, ctx=Store())], value=val)
        body.append(line)
    #
    for n, (lhs, neq) in enumerate(fff.content.items()):
        line = Assign(targets=[Subscript(value=Name(id='out', ctx=Load()),
                                         slice=Index(Num(n)), ctx=Store())], value=Name(id=lhs, ctx=Load()))
        body.append(line)

    if sys.version_info >= (3, 8, 0):
        f = FunctionDef(name=funname,
                        args=ast_arguments(posonlyargs=[], args=[arg(arg=a) for a in arguments] + [arg(arg='out')],
                                           vararg=None, kwarg=None, kwonlyargs=[], kw_defaults=[], defaults=[]),
                        body=unpacking + body, decorator_list=[])
        mod = Module(body=[f], type_ignores=[])
    else:
        f = FunctionDef(name=funname,
                        args=ast_arguments(args=[arg(arg=a) for a in arguments] + [arg(arg='out')],
                                           vararg=None, kwarg=None, kwonlyargs=[], kw_defaults=[], defaults=[]),
                        body=unpacking + body, decorator_list=[])
        mod = Module(body=[f])

    mmod = ast.fix_missing_locations(mod)
    return mmod

def make_method_from_factory(fff: FlatFunctionFactory, vectorize=True, use_file=False, debug=False):
    mod = compile_factory(fff)

    if debug:
        from .codegen import to_source
        print(to_source(mod))

    if vectorize:
        coredims = [len(v) for k, v in fff.arguments.items()]
        signature = str.join(',', ['(n_{})'.format(d) for d in coredims])
        n_out = len(fff.content)
        if n_out in coredims:
            signature += '->(n_{})'.format(n_out)
            # ftylist = float64[:](*([float64[:]] * len(coredims)))
            fty = "void(*[float64[:]]*{})".format(len(coredims) + 1)
        else:
            signature += ',(n_{})'.format(n_out)
            fty = "void(*[float64[:]]*{})".format(len(coredims) + 1)
    else:
        signature = None

    fun = eval_ast(mod)

    from numba import jit, guvectorize

    jitted = jit(fun, nopython=True)
    if vectorize:
        gufun = guvectorize(
            [fty], signature, target='parallel', nopython=True)(fun)
        return jitted, gufun
    else:
        return jitted


def eval_ast(mod):
    context = {}

    import numpy

    context['inf'] = numpy.inf

    context['exp'] = numpy.exp
    context['log'] = numpy.log
    context['sin'] = numpy.sin
    context['cos'] = numpy.cos
    context['tan'] = numpy.tan
    context['tanh'] = numpy.tanh
    context['atan'] = numpy.arctan
    context['atanh'] = numpy.arctanh

    context['maximum'] = numpy.maximum
    context['minimum'] = numpy.minimum
    context['max'] = numpy.maximum
    context['min'] = numpy.minimum
    context['abs'] = numpy.abs

    name = mod.body[0].name
    mod = ast.fix_missing_locations(mod)
    code = compile(mod, '<string>', 'exec')
    exec(code, context, context)
    fun = context[name]

    return fun



def test_compiler():
    from dolang.factory import FlatFunctionFactory
    from dolang.codegen import to_source

    fff = FlatFunctionFactory(
        dict( # preamble
            x="b+c*p1",
            y="(a+b)*c"
        ),
        dict(
            out_1="exp(y*p1)-a*p2",
            out_2="x+y"
        ),
        dict(
            g1=['a','b'],
            g2=['c'],
            g3=['p1','p2']
        ),
        'testfun'
    )

    from dolang.function_compiler import make_method_from_factory

    fun = make_method_from_factory(fff)[0]

    import numpy as np
    out = np.array([0.3, 0.1])

    fun(np.array([0.1, 0.2]), np.array([10]), np.array([0.5, 0.3]), out)

    assert( abs(out[0] - 4.45168907)<1e-8 )
    assert( abs(out[1] - 8.2)<1e-8 )





from .grammar import *
parse_string = parser.parse


import ast

def eval_scalar(tree):
    try:
        if isinstance(tree, ast.Num):
            return tree.n
        elif isinstance(tree, ast.UnaryOp):
            if isinstance(tree.op, ast.USub):
                return -tree.operand.n
            elif isinstance(tree.op, ast.UAdd):
                return tree.operand.n
        else:
            raise Exception("Don't know how to do that.")
    except:
        raise Exception("Don't know how to do that.")


# from dolang.language import functions as functions_dict
# from dolang.codegen import to_source
# from typing import Tuple, List, Dict, Set, TypeVar, Union
# from ast import NodeTransformer, Name, UnaryOp, UAdd, USub, Load, Call

# try:
#     from .dataclasses import dataclass
# except:
#     from dataclasses import dataclass

# import ast
# from ast import Expr
# from functools import wraps
# import re

# T = TypeVar('T')

# Expression = ast.Expr

# functions = list(functions_dict.keys())

# __regex_eq__ = re.compile("([^=]*)=([^=]*)")


# def parse_string(s: str) -> ast.Expr:
#     ss = s.replace("^", "**")
#     m = re.match(__regex_eq__, ss)
#     if m:
#         ss = str.join('==', m.groups())
#     return ast.parse(ss).body[0]


# def dedup(l: List[T]) -> List[T]:
#     return list(dict.fromkeys(l))


# def stringify_variable(arg: Tuple[str, int]) -> str:
#     s = arg[0]
#     date = arg[1]
#     if date == 0:
#         return '{}__0_'.format(s)
#     elif date <= 0:
#         return '{}_m{}_'.format(s, str(-date))
#     elif date >= 0:
#         return '{}__{}_'.format(s, str(date))


# def stringify_parameter(p: str) -> str:
#     return '{}_'.format(p)


# def stringify_symbol(arg) -> str:
#     if isinstance(arg, str):
#         return stringify_parameter(arg)
#     elif isinstance(arg, tuple):
#         if len(arg) == 2 and isinstance(arg[0], str) and isinstance(
#                 arg[1], int):
#             return stringify_variable(arg)
#     raise Exception("Unknown canonical form: {}".format(arg))


# def destringify_variable(s: str) -> Tuple[str, int]:
#     raise Exception("Not implemented.")


# def destringify_parameter(s: str) -> Tuple[str, int]:
#     raise Exception("Not implemented.")


# def destringify(s: str):
#     raise Exception("Not implemented.")


# ###
# # The following function can take a string or an expression as the first argument, and return an argument of the same type.
# ###


# def expression_or_string(f):
#     @wraps(f)
#     def wrapper(*args, **kwds):
#         if not isinstance(args[0], str):
#             return f(*args, **kwds)
#         else:
#             a = parse_string(args[0])
#             nargs = tuple([a]) + args[1:]
#             res = f(*nargs, **kwds)
#             return to_source(res)

#     return wrapper


# @expression_or_string
# def stringify(expr: Expression, variables: List[str] = []) -> Expression:
#     import copy
#     en = ExpressionStringifier(variables=variables)
#     return en.visit(copy.deepcopy(expr))


# # shift timing in equations
# # time_shift(:(a+b(1)+c),1,[:b,:c]) == :(a+b(2)+c(1))


# @expression_or_string
# def time_shift(expr: Expression, n) -> Expression:
#     import copy
#     eexpr = copy.deepcopy(expr)
#     variables = [e[0] for e in list_variables(eexpr)]
#     return TimeShiftTransformer(shift=n, variables=variables).visit(eexpr)


# @expression_or_string
# def steady_state(expr: Expression) -> Expression:
#     import copy
#     eexpr = copy.deepcopy(expr)
#     variables = [e[0] for e in list_variables(eexpr)]
#     return TimeShiftTransformer(shift='S', variables=variables).visit(eexpr)


# @expression_or_string
# def sanitize(expr: Expression, variables=None, functions=None):
#     # special functions ?
#     es = ExpressionSanitizer(variables=variables)
#     return es.visit(expr)


# def list_variables(expr: Expression,
#                    funs: List[str] = None) -> List[Tuple[str, int]]:

#     if funs is None:
#         funs = []
#     ll = ListSymbols(known_functions=functions + funs)
#     ll.visit(expr)
#     if ll.problems:
#         e = Exception('Symbolic error.')
#         e.problems = ll.problems
#         raise e
#     return dedup([v[0] for v in ll.variables])


# @dataclass
# class SymbolList(dict):
#     variables: Set[Tuple[str, int]]
#     parameters: Set[str]
#     functions: Set[str]


# def list_symbols(expr: Expression, funs: List[str] = None) -> SymbolList:
#     if funs is None:
#         funs = []
#     ll = ListSymbols(known_functions=functions + funs)
#     ll.visit(expr)
#     if ll.problems:
#         e = Exception('Symbolic error.')
#         e.problems = ll.problems
#         print(e.problems)
#         raise e

#     def head(v):
#         return [i[0] for i in v]

#     return SymbolList(
#         dedup(head(ll.variables)), dedup(head(ll.constants)),
#         dedup(head(ll.functions)))


# ### Tree manipulation functions


# def eval_scalar(tree: Expression) -> Union[int, float]:
#     try:
#         if isinstance(tree, ast.Num):
#             return tree.n
#         elif isinstance(tree, ast.UnaryOp):
#             if isinstance(tree.op, ast.USub):
#                 return -tree.operand.n
#             elif isinstance(tree.op, ast.UAdd):
#                 return tree.operand.n
#         else:
#             raise Exception("Don't know how to do that.")
#     except:
#         raise Exception("Don't know how to do that.")


# class ListSymbols(ast.NodeVisitor):
#     def __init__(self, known_functions=[], known_variables=[]):
#         self.known_functions = known_functions
#         self.known_variables = known_variables
#         self.functions = []
#         self.variables = []
#         self.constants = []
#         self.problems = []

#     def visit_Call(self, call):
#         name = call.func.id
#         colno = call.func.col_offset
#         if name in self.known_functions:
#             self.functions.append((name, colno))
#             [self.visit(e) for e in call.args]
#         else:
#             try:
#                 assert (len(call.args) == 1)
#                 n = int(eval_scalar(call.args[0]))
#                 self.variables.append(((name, n), colno))
#             except Exception as e:
#                 self.problems.append([name, 0, colno, 'incorrect subscript'])
#                 # [self.visit(e) for e in call.args]

#     def visit_Name(self, name):
#         # colno = name.colno
#         colno = name.col_offset
#         n = 0
#         name = name.id
#         if name in self.known_functions:
#             self.problems.append([name, colno, 'function_not_called'])
#         else:
#             self.constants.append((name, colno))


# class ExpressionStringifier(NodeTransformer):

#     # replaces calls to variables by time subscripts

#     def __init__(self, variables=None, functions=None, constants=None):

#         self.variables = variables if variables is not None else []
#         if functions is None:
#             self.functions = [e for e in functions_dict.keys()]
#         else:
#             self.functions = functions
#         if constants is None:
#             self.constants = ['pi', 'e', 'inf']
#         else:
#             self.constants = constants
#         # self.variables = tvariables # ???

#     def visit_Name(self, node):

#         name = node.id
#         # if name self.functions:
#         #     return node
#         if name in self.constants:
#             return node
#         elif name in self.variables:
#             return Name(id=stringify_variable((name, 0)), ctx=Load())
#         else:
#             return Name(id=stringify_parameter(name), ctx=Load())

#     def visit_Call(self, node):

#         name = node.func.id
#         args = node.args[0]

#         if name in self.variables or name not in self.functions:
#             try:
#                 date = eval_scalar(args)
#             except:
#                 raise Exception("Unrecognized subscript.")
#             newname = stringify_variable((name, date))
#             if newname is not None:
#                 return Name(newname, Load())
#         else:
#             return Call(
#                 func=node.func,
#                 args=[self.visit(e) for e in node.args],
#                 keywords=[])


# class TimeShiftTransformer(ast.NodeTransformer):
#     def __init__(self, shift=0, variables=None):

#         self.variables = variables if variables is not None else []
#         self.shift = shift

#     def visit_Call(self, node):

#         name = node.func.id
#         args = node.args[0]

#         if name in self.variables:
#             if isinstance(args, UnaryOp):
#                 # we have s(+1)
#                 if (isinstance(args.op, UAdd)):
#                     args = args.operand
#                     date = args.n
#                 elif (isinstance(args.op, USub)):
#                     args = args.operand
#                     date = -args.n
#                 else:
#                     raise Exception("Unrecognized subscript.")
#             else:
#                 date = args.n
#             if self.shift == 'S':
#                 new_date = 0
#             else:
#                 new_date = date + self.shift
#             return ast.parse('{}({})'.format(name, new_date)).body[0].value
#         else:

#             # , keywords=node.keywords,  kwargs=node.kwargs)
#             return Call(
#                 func=node.func,
#                 args=[self.visit(e) for e in node.args],
#                 keywords=[])


# class ExpressionSanitizer(NodeTransformer):

#     # replaces calls to variables by time subscripts
#     def __init__(self, variables=None):
#         self.variables = variables if variables is not None else []

#     def visit_Name(self, node):
#         name = node.id
#         if name in self.variables:
#             return ast.parse('{}(0)'.format(name)).body[0].value
#         else:
#             return node

#     def visit_Call(self, node):
#         name = node.func.id
#         if name in self.variables:
#             t = eval_scalar(node.args[0])
#             return ast.parse('{}({})'.format(name, t)).body[0].value
#         else:
#             return Call(
#                 func=node.func,
#                 args=[self.visit(e) for e in node.args],
#                 keywords=[])


# class NameSubstituter(ast.NodeTransformer):

#     # substitutes a name by an expression
#     def __init__(self, substitutions: Dict[str, Expr]):
#         self.substitutions = substitutions

#     def visit_Name(self, node):
#         name = node.id
#         if name in self.substitutions:
#             return self.substitutions[name]
#         else:
#             return node



import lark
from lark.tree import Tree
from lark.lexer import Token
from lark.visitors import Interpreter,  Visitor, Transformer

from functools import wraps

from typing import Tuple, Dict, Set, Union, List
Expression = Union[Tree, Token]

from dataclasses import dataclass

## the following grammar represents equations
## it recognizes variables indexed as v[t] or v[t+k] where k is an integer
## for compatibility purposes, it also recognizes v(0) as v[t] unless 
## v is a prespecified function.
## later, this compatibility feature will be turned off.

grammar_0 = """
    ?start: equation

    ?equation: double_complementarity | equality | formula
    equality: sum "=" sum 
        | sum "==" sum
    double_inequality: formula "<=" symbol "<=" formula
        | formula "<=" variable "<=" formula
    double_complementarity: formula _PERP double_inequality

    _PERP: "⟂" | "|"

    ?formula: sum
    ?sum: product
        | sum "+" product   -> add
        | sum "-" product   -> sub
    ?product: atom
        | product "*" atom  -> mul
        | product "/" atom  -> div
    ?pow: atom "^" atom
    ?atom: NUMBER           -> number
         | "-" atom         -> neg
         | pow
         | symbol            
         | "(" sum ")"
         | call
         | variable

    !symbol: NAME -> symbol
    variable: cname  "[" date_index "]" -> variable
            | cname "[" "t" "]" -> variable
            | cname "(" date_index ")" -> variable
    !cname: NAME -> name
    ?date_index: "t" SIGNED_INT2 -> date
        | SIGNED_INT -> date
    ?call: FUNCTION "(" sum ")" -> call
    FUNCTION: "sin"|"cos"|"exp"|"log"

    SIGNED_INT2: ("+"|"-") INT
    %import common.SIGNED_INT
    %import common.INT
    %import common.CNAME -> NAME
    %import common.NUMBER
    %import common.WS_INLINE
    %ignore WS_INLINE
"""

from lark.lark import Lark
parser = Lark(grammar_0)


# Prints a tree as a string
# WIP!!! (probably correct, but way too many parentheses)
class Printer(Interpreter):
    def add(self, tree):
        if len(tree.children) ==1:
            return "ERROR"
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} + {b}"
    def sub(self, tree):
        if len(tree.children) ==1:
            return "ERROR"
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} - ({b})"
    def variable(self, tree):
        name = tree.children[0].children[0].value
        try:
            time = int(tree.children[1].children[0].value)
        except:
            time = 0
        if time ==0:
            ds = "t"
        elif time>0:
            ds = "t+"+str(time)
        elif time<0:
            ds = "t-"+str(-time)
        return f"{name}[{ds}]"

    def equality(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} = {b}"

    def double_complementarity(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} ⟂ {b}"

    def double_inequality(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        c = self.visit(tree.children[2])
        return f"{a} <= {b} <= {c}"

    def symbol(self, tree):
        name = tree.children[0].value
        return (name)
    def mul(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"({a})*({b})"
    def div(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"({a})/({b})"
    def call(self, tree):
        funname = tree.children[0].value
        args = self.visit( tree.children[1] )
        return f"{funname}({args})"
    def pow(self, tree):
        arg1 = self.visit(tree.children[0])
        arg2 = self.visit(tree.children[1])
        return f"({arg1})^({arg2})"
    def number(self, tree):
        return tree.children[0].value
    def signed_int(self, tree):
        return tree.children[0].value
    def neg(self, tree):
        a = self.visit(tree.children[0])
        return f"-({a})"




## replaces v[t] by v[t+0] (I didn't find how to do it in the grammar)
## replaces v by v[t] when v identified as a variable
class Sanitizer(Transformer):

    def __init__(self, variables=[]):
        self.__variables__= variables

    def symbol(self, *args):
        tok = args[0][0]
        val = tok.value
        if val in self.__variables__:
            return Tree("variable", [Tree("name", [Token("NAME", val)]), Tree("date", [Token("NUMBER", '0')])])
        else:
            return Tree("symbol", *args)

    def variable(self, *args):
        if len(args[0])==1:
            date = Tree('date',[Token("NUMBER", '0')])
            args = (args[0] + [date], )
        return Tree("variable", *args)

def stringify_variable(arg: Tuple[str, int]) -> str:
    s = arg[0]
    date = arg[1]
    if date == 0:
        return '{}__0_'.format(s)
    elif date <= 0:
        return '{}_m{}_'.format(s, str(-date))
    elif date >= 0:
        return '{}__{}_'.format(s, str(date))


def stringify_parameter(p: str) -> str:
    return '{}_'.format(p)


def stringify_symbol(arg) -> str:
    if isinstance(arg, str):
        return stringify_parameter(arg)
    elif isinstance(arg, tuple):
        if len(arg) == 2 and isinstance(arg[0], str) and isinstance(
                arg[1], int):
            return stringify_variable(arg)
    raise Exception("Unknown canonical form: {}".format(arg))



## replaces symbols and variables by their canonical string represantation:
## symbols s -> "s_"
## variable v(1) -> "v_p1_"
## variable v(-1) -> "v_m1_"
class Stringifier(Transformer):

    def symbol(self, children):
        
        name = children[0].value
        if name == "inf":
            s = name
        else:
            s = stringify_parameter( name)
        return Tree("symbol", [Token("NAME",s)])

    def variable(self, children):
        
        name = children[0].children[0].value
        if len(children) == 1:
            date = 0
        else:
            date = int(children[1].children[0].value)
        s = stringify_variable( (name, date) )
        return Tree("symbol", [Token("NAME",s)])


###
## if shift == 'S' replaces all v[t+k] by v[t] (is that reasonable ?)
## if shift is an integer, replaces v[t+k] by v[t+k+shift]
class TimeShifter(Transformer):

    def __init__(self, shift: int):
        self.shift = shift

    def variable(self, children):
        
        name = children[0].children[0].value
        try:
            date = int(children[1].children[0].value)
        except:
            date = 0
        if self.shift=="S":
            new_date = "0"
        else:
            new_date = str(date + self.shift)
        return Tree("variable", [Tree("name", [Token("NAME", name)]), Tree("date", [Token("NUMBER", new_date)])])



@dataclass
class SymbolList(dict):
    variables: List[Tuple[str, int]]
    parameters: List[str]
    functions: List[str]

## lists all variables in an  expression
## result field contains variables with timing, parameters, and functions used
class VariablesLister(Visitor):

    def __init__(self):
        self.result = SymbolList([],[],[])

    def variable(self, tree):
        children = tree.children
        
        name = children[0].children[0].value
        date = int(children[1].children[0].value)
        if (name,date) not in self.result.variables:
            self.result.variables.append((name, date))

    def symbol(self, tree):
        children = tree.children
        name = children[0].value
        if name not in self.result.parameters:
            self.result.parameters.append(name)

    def call(self, tree):
        children = tree.children
        name = children[0].value
        if name not in self.result.functions:
            self.result.functions.append(name)


## replaces symbols in an expression:
## NameSustituter({'h': parser.parse("e+1")}).visit(parser.parse("h+p"))  returns 
## parser.parse("e+1+p")
class NameSubstituter(Transformer):

    # substitutes a symbol by an expression
    def __init__(self, substitutions: Dict[str, Expression]):
        self.substitutions = substitutions

    def symbol(self, children):
        name = children[0].value
        if name in self.substitutions:
            return self.substitutions[name]
        else:
            return Tree("symbol", children)

def subs(expr: str, substitutions):
    import copy
    s = dict()
    f = expr
    for k,v in substitutions.items():
        s[k] = parser.parse(v)
    ns = NameSubstituter(s)
    res = ns.transform(f)
    return str_expression(res)

# prints expression
def str_expression(expr: Expression)->str:
    return Printer().visit(expr)


# decorator to define functions which operate
# either on Trees or on strings.
def expression_or_string(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        if not isinstance(args[0], str):
            return f(*args, **kwds)
        else:
            a = parser.parse(args[0])
            nargs = tuple([a]) + args[1:]
            res = f(*nargs, **kwds)
            return str_expression(res)

    return wrapper


## these functions apply the vistors/transformers either on expressions or on strings 
@expression_or_string
def stringify(expr: Expression):
    return Stringifier().transform(expr)

@expression_or_string
def time_shift(expr: Expression, n) -> Expression:

    return TimeShifter(n).transform(expr)


@expression_or_string
def steady_state(expr: Expression) -> Expression:
    return TimeShifter(shift='S').transform(expr)


@expression_or_string
def sanitize(expr: Expression, variables=[]):
    return Sanitizer(variables=variables).transform(expr)
    


def list_symbols(expr: Union[Expression,str]) -> SymbolList:

    if isinstance(expr, str):
        expr = parser.parse(str)

    ll = VariablesLister()
    ll.visit(expr)

    return ll.result

def list_variables(expr: Union[Expression, str]):
    return list_symbols(expr).variables


def test():

    l = lark.Lark(grammar_0)

    s = "d + h + oij^2 - (1.0 + a) + sin(x) + v[t+1] + v[t] + x(0) + z[t]"
    p0 = l.parse(s)

    list_variables = ['h']
    p = Sanitizer(list_variables).transform(p0)

    pp = time_shift(s,-2)
    pp_2 = time_shift(s, "S")

    print("Original:                 ", s)
    print("Steady-state:             ", steady_state(s))
    print("Sanitize (variables=[h]): ", sanitize(s, variables=['h']))
    print("Stringify: ", stringify(s))

    print("Original Tree:")
    print( p0.pretty() )

if __name__ == "__main__":
    test()


def tree_to_ast(tree):
    import ast
    import re
    __regex_eq__ = re.compile("([^=]*)=([^=]*)")

    txt = (str_expression(tree))

    ss = txt.replace("^", "**")
    m = re.match(__regex_eq__, ss)
    if m:
        ss = str.join('==', m.groups())
    
    return ast.parse(ss).body[0]



import ast

class Compare:

    def __init__(self):
        self.d = {}

    def compare(self, A, B):
        if isinstance(A, ast.Name) and (A.id[0] == '_'):
            if A.id not in self.d:
                self.d[A.id] = B
                return True
            else:
                return self.compare(self.d[A.id], B)
        if not (A.__class__ == B.__class__): return False
        if isinstance(A, ast.Name):
            return A.id == B.id
        elif isinstance(A, ast.Call):
            if not self.compare(A.func, B.func): return False
            if not len(A.args)==len(B.args): return False
            for i in range(len(A.args)):
                if not self.compare(A.args[i], B.args[i]): return False
            return True
        elif isinstance(A, ast.Num):
            return A.n == B.n
        elif isinstance(A, ast.Expr):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Module):
            if not len(A.body)==len(B.body): return False
            for i in range(len(A.body)):
                if not self.compare(A.body[i], B.body[i]): return False
            return True
        elif isinstance(A, ast.BinOp):
            if not isinstance(A.op, B.op.__class__): return False
            if not self.compare(A.left, B.left): return False
            if not self.compare(A.right, B.right): return False
            return True
        elif isinstance(A, ast.UnaryOp):
            if not isinstance(A.op, B.op.__class__): return False
            return self.compare(A.operand, B.operand)
        elif isinstance(A, ast.Subscript):
            if not self.compare(A.value, B.value): return False
            return self.compare(A.slice, B.slice)
        elif isinstance(A, ast.Index):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Compare):
            if not self.compare(A.left, B.left): return False
            if not len(A.ops)==len(B.ops): return False
            for i in range(len(A.ops)):
                if not self.compare(A.ops[i], B.ops[i]): return False
            if not len(A.comparators)==len(B.comparators): return False
            for i in range(len(A.comparators)):
                if not self.compare(A.comparators[i], B.comparators[i]): return False
            return True
        elif isinstance(A, ast.In):
            return True
        elif isinstance(A, (ast.Eq, ast.LtE)):
            return True
        else:
            print(A.__class__)
            raise Exception("Not implemented")


def compare_strings(a,b):
    t1 = ast.parse(a)
    t2 = ast.parse(b)
    comp = Compare()
    val = comp.compare(t1,t2)
    d = comp.d
    return val

def match(m,s):
    if isinstance(m,str):
        m = ast.parse(m).body[0].value
    if isinstance(s,str):
        s = ast.parse(s).body[0].value
    comp = Compare()
    val = comp.compare(m,s)
    d = comp.d
    if len(d) == 0:
        return val
    else:
        return d

import ast
class ReplaceExpectation(ast.NodeTransformer):
    import ast
    def visit_Subscript(self, node):
        from ast import Expr, Call, Name, Load, Lambda, arguments, arg
        m = match('E[ _expr | _x in _set]', node)
        if m:
            x_s = m['_x'].id # name of dummy vaqriable
            expr = m['_expr']
            sset = m['_set']
            res = Call(func=Name(id='Ex', ctx=Load()),
                    args=[Lambda(args=arguments(args=[arg(arg=x_s, annotation=None)],
                    vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=expr), sset],
                    keywords=[], starargs=None, kwargs=None)
            return res
        m = match('Sum[ _expr | _x in _set]', node)
        if m:
            x_s = m['_x'].id # name of dummy vaqriable
            expr = m['_expr']
            sset = m['_set']
            res = Call(func=Name(id='Sum', ctx=Load()),
                    args=[Lambda(args=arguments(args=[arg(arg=x_s, annotation=None)],
                    vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=expr), sset],
                    keywords=[], starargs=None, kwargs=None)
            return res

        return node




def test_strings():

    tests = [
    ['b(_x)','b(1)', True],
    ['_x(1)','b(1)', True],
    ['_x(2)','b(1)', False],
    ['a + b + c', 'a + b + c', True],
    ['a + b + d', 'a + b + c', False],
    ['a + b + _d','a + b + c',  True],
    ['a + _d + _d','a + b + c', False],
    ['a + _d + _d','a + c + c', True],
    ['a*_x+c*_x**2+d*_y',  'a*x+c*x**2+d*y', True],
    ['a*_x+c*_x**2+d*_y',  'a*x+c*y**2+d*z', False],
    ['-_a(_x) + _b(_x)',  '-u(-1)+v(-1)', True],
    ['-_a(_x) + _b(_x)',  '-u(11)+v(11)', True],
    ['-_a(_x) + _b(_x)',  '-u(1)+v(2)', False],
    ['E[_x]',  'E[1]', True],
    ['E[2]',  'E[1]', False],
    ['E[1]',  'X[1]', False],
    ['a in _x',  'a in b', True],
    ['a in b',  'a in c', False],
    ['a | _x',  'a | c', True],
    ['a | b',  'a | c', False],
    ['E[ _x | _y in _z]',  'E[ a | b in c]', True],
    ['E[ a | b in c]',  'E[ a | b in d]', False],
    ['E[ _x | _y in _z]',  'E[ f(a) | b in c]', True],
    ['E[ _x | _y in _z]',  'E[ f(a)^2 | b in c]', True],
    # ['E[ f(s) ]',  'E[ _x ]', True],
    ['_a == _b | _x <= _y <= _z', 'a == b | x <= y <= z', True],
    ['_eq | _comp','phi[s] + (-(f[s]-min_f))*1000    |  phi[s]', True],
    ['_eq | _comp','phi[s] + (-(f[s]-min_f))*1000    | 0 <= phi[s]', True],
    ['_eq | 0 <= _x','phi[s] + (-(f[s]-min_f))*1000  | 0 <= phi[s]', True],
    # ['_cond : _eq | _comp' ,  's>0 : a == 8 | x<=0<=y', True],
    ]
    replacement_rules = [
        ['E[ expr | x in S(s)]', 'Esp(lambda x: expr, S(s))'],
        ['Sum[ expr | x in H(s)]', 'Sum(lambda x: expr, H(s)']
    ]

    for l in tests:
        print('{} : {} : {} : {}'.format(compare_strings(l[0], l[1]), l[2], l[0],l[1]))
        # assert(compare_strings(l[0], l[1]) ==  l[2])


def fix_equation(eq):
    eqq = ReplaceExpectation().visit(eq)
    return eqq
#
#
# if __name__ == "__main__":
#     test_strings()
#     res = match("E[ _i | _v in _s]", "E[ y(1)*2 | y(1) in S(s)]")
#     res = match('_a == _b | _x <= _y <= _z', 'a == b | x <= y <= z')
#     if match("a_","b"):
#         print("yes")
#     if match("a","a"):
#         print("yes")
#     else:
#         print("no")



def write_time_subscript(s):

    neg = 'm'
    if isinstance(s, int):
        if s>=0:
            return str(s)
        elif s<0:
            return neg + str(-s)
    else:
        raise Exception("Don't know what to do with that")

class Symbol:

    def __init__(self, name=None, subscripts=None, date=None):
        if date is None:
            date = 0
        if subscripts is None:
            subscripts = []
        assert(isinstance(date,int))
        for s in subscripts:
            assert(isinstance(s,str))
            assert('__' not in s)
        assert(name.isidentifier())
        self.subscripts = tuple( subscripts )
        self.date = date
        self.name = name

    def __str__(self):
        # sep = 'ⵂ'
        # close_lhs = 'ⵂ'
        # close_rhs = 'ⵂ'
        sep = '__'
        close_lhs = '__'
        close_rhs = ''
        # close_lhs = '᚜'
        # close_rhs = '᚛'

        if self.date:
            argss = str.join(sep, [write_time_subscript(self.date)] + list(self.subscripts))
        else:
            argss =  str.join(sep, list(self.subscripts))
        s = self.name + close_lhs + argss + close_rhs
        return s



import sys
import types
from copy import deepcopy
import inspect

__all__ = ['dataclass',
           'field',
           'FrozenInstanceError',
           'InitVar',
           'MISSING',

           # Helper functions.
           'fields',
           'asdict',
           'astuple',
           'make_dataclass',
           'replace',
           'is_dataclass',
           ]

# Conditions for adding methods.  The boxes indicate what action the
#  dataclass decorator takes.  For all of these tables, when I talk
#  about init=, repr=, eq=, order=, hash=, or frozen=, I'm referring
#  to the arguments to the @dataclass decorator.  When checking if a
#  dunder method already exists, I mean check for an entry in the
#  class's __dict__.  I never check to see if an attribute is defined
#  in a base class.

# Key:
# +=========+=========================================+
# + Value   | Meaning                                 |
# +=========+=========================================+
# | <blank> | No action: no method is added.          |
# +---------+-----------------------------------------+
# | add     | Generated method is added.              |
# +---------+-----------------------------------------+
# | add*    | Generated method is added only if the   |
# |         |  existing attribute is None and if the  |
# |         |  user supplied a __eq__ method in the   |
# |         |  class definition.                      |
# +---------+-----------------------------------------+
# | raise   | TypeError is raised.                    |
# +---------+-----------------------------------------+
# | None    | Attribute is set to None.               |
# +=========+=========================================+

# __init__
#
#   +--- init= parameter
#   |
#   v     |       |       |
#         |  no   |  yes  |  <--- class has __init__ in __dict__?
# +=======+=======+=======+
# | False |       |       |
# +-------+-------+-------+
# | True  | add   |       |  <- the default
# +=======+=======+=======+

# __repr__
#
#    +--- repr= parameter
#    |
#    v    |       |       |
#         |  no   |  yes  |  <--- class has __repr__ in __dict__?
# +=======+=======+=======+
# | False |       |       |
# +-------+-------+-------+
# | True  | add   |       |  <- the default
# +=======+=======+=======+


# __setattr__
# __delattr__
#
#    +--- frozen= parameter
#    |
#    v    |       |       |
#         |  no   |  yes  |  <--- class has __setattr__ or __delattr__ in __dict__?
# +=======+=======+=======+
# | False |       |       |  <- the default
# +-------+-------+-------+
# | True  | add   | raise |
# +=======+=======+=======+
# Raise because not adding these methods would break the "frozen-ness"
#  of the class.

# __eq__
#
#    +--- eq= parameter
#    |
#    v    |       |       |
#         |  no   |  yes  |  <--- class has __eq__ in __dict__?
# +=======+=======+=======+
# | False |       |       |
# +-------+-------+-------+
# | True  | add   |       |  <- the default
# +=======+=======+=======+

# __lt__
# __le__
# __gt__
# __ge__
#
#    +--- order= parameter
#    |
#    v    |       |       |
#         |  no   |  yes  |  <--- class has any comparison method in __dict__?
# +=======+=======+=======+
# | False |       |       |  <- the default
# +-------+-------+-------+
# | True  | add   | raise |
# +=======+=======+=======+
# Raise because to allow this case would interfere with using
#  functools.total_ordering.

# __hash__

#      +------------------- hash= parameter
#      |       +----------- eq= parameter
#      |       |       +--- frozen= parameter
#      |       |       |
#      v       v       v    |        |        |
#                           |   no   |  yes   |  <--- class has __hash__ in __dict__?
# +=========+=======+=======+========+========+
# | 1 None  | False | False |        |        | No __eq__, use the base class __hash__
# +---------+-------+-------+--------+--------+
# | 2 None  | False | True  |        |        | No __eq__, use the base class __hash__
# +---------+-------+-------+--------+--------+
# | 3 None  | True  | False | None   |        | <-- the default, not hashable
# +---------+-------+-------+--------+--------+
# | 4 None  | True  | True  | add    | add*   | Frozen, so hashable
# +---------+-------+-------+--------+--------+
# | 5 False | False | False |        |        |
# +---------+-------+-------+--------+--------+
# | 6 False | False | True  |        |        |
# +---------+-------+-------+--------+--------+
# | 7 False | True  | False |        |        |
# +---------+-------+-------+--------+--------+
# | 8 False | True  | True  |        |        |
# +---------+-------+-------+--------+--------+
# | 9 True  | False | False | add    | add*   | Has no __eq__, but hashable
# +---------+-------+-------+--------+--------+
# |10 True  | False | True  | add    | add*   | Has no __eq__, but hashable
# +---------+-------+-------+--------+--------+
# |11 True  | True  | False | add    | add*   | Not frozen, but hashable
# +---------+-------+-------+--------+--------+
# |12 True  | True  | True  | add    | add*   | Frozen, so hashable
# +=========+=======+=======+========+========+
# For boxes that are blank, __hash__ is untouched and therefore
#  inherited from the base class.  If the base is object, then
#  id-based hashing is used.
# Note that a class may have already __hash__=None if it specified an
#  __eq__ method in the class body (not one that was created by
#  @dataclass).


# Raised when an attempt is made to modify a frozen class.
class FrozenInstanceError(AttributeError): pass

# A sentinel object for default values to signal that a
#  default-factory will be used.
# This is given a nice repr() which will appear in the function
#  signature of dataclasses' constructors.
class _HAS_DEFAULT_FACTORY_CLASS:
    def __repr__(self):
        return '<factory>'
_HAS_DEFAULT_FACTORY = _HAS_DEFAULT_FACTORY_CLASS()

# A sentinel object to detect if a parameter is supplied or not.  Use
#  a class to give it a better repr.
class _MISSING_TYPE:
    pass
MISSING = _MISSING_TYPE()

# Since most per-field metadata will be unused, create an empty
#  read-only proxy that can be shared among all fields.
_EMPTY_METADATA = types.MappingProxyType({})

# Markers for the various kinds of fields and pseudo-fields.
_FIELD = object()                 # An actual field.
_FIELD_CLASSVAR = object()        # Not a field, but a ClassVar.
_FIELD_INITVAR = object()         # Not a field, but an InitVar.

# The name of an attribute on the class where we store the Field
#  objects. Also used to check if a class is a Data Class.
_MARKER = '__dataclass_fields__'

# The name of the function, that if it exists, is called at the end of
# __init__.
_POST_INIT_NAME = '__post_init__'


class _InitVarMeta(type):
    def __getitem__(self, params):
        return self

class InitVar(metaclass=_InitVarMeta):
    pass


# Instances of Field are only ever created from within this module,
#  and only from the field() function, although Field instances are
#  exposed externally as (conceptually) read-only objects.
# name and type are filled in after the fact, not in __init__. They're
#  not known at the time this class is instantiated, but it's
#  convenient if they're available later.
# When cls._MARKER is filled in with a list of Field objects, the name
#  and type fields will have been populated.
class Field:
    __slots__ = ('name',
                 'type',
                 'default',
                 'default_factory',
                 'repr',
                 'hash',
                 'init',
                 'compare',
                 'metadata',
                 '_field_type',  # Private: not to be used by user code.
                 )

    def __init__(self, default, default_factory, init, repr, hash, compare,
                 metadata):
        self.name = None
        self.type = None
        self.default = default
        self.default_factory = default_factory
        self.init = init
        self.repr = repr
        self.hash = hash
        self.compare = compare
        self.metadata = (_EMPTY_METADATA
                         if metadata is None or len(metadata) == 0 else
                         types.MappingProxyType(metadata))
        self._field_type = None

    def __repr__(self):
        return ('Field('
                f'name={self.name!r},'
                f'type={self.type},'
                f'default={self.default},'
                f'default_factory={self.default_factory},'
                f'init={self.init},'
                f'repr={self.repr},'
                f'hash={self.hash},'
                f'compare={self.compare},'
                f'metadata={self.metadata}'
                ')')


# This function is used instead of exposing Field creation directly,
#  so that a type checker can be told (via overloads) that this is a
#  function whose type depends on its parameters.
def field(*, default=MISSING, default_factory=MISSING, init=True, repr=True,
          hash=None, compare=True, metadata=None):
    """Return an object to identify dataclass fields.

    default is the default value of the field. default_factory is a
    0-argument function called to initialize a field's value. If init
    is True, the field will be a parameter to the class's __init__()
    function. If repr is True, the field will be included in the
    object's repr(). If hash is True, the field will be included in
    the object's hash(). If compare is True, the field will be used in
    comparison functions. metadata, if specified, must be a mapping
    which is stored but not otherwise examined by dataclass.

    It is an error to specify both default and default_factory.
    """

    if default is not MISSING and default_factory is not MISSING:
        raise ValueError('cannot specify both default and default_factory')
    return Field(default, default_factory, init, repr, hash, compare,
                 metadata)


def _tuple_str(obj_name, fields):
    # Return a string representing each field of obj_name as a tuple
    #  member. So, if fields is ['x', 'y'] and obj_name is "self",
    #  return "(self.x,self.y)".

    # Special case for the 0-tuple.
    if not fields:
        return '()'
    # Note the trailing comma, needed if this turns out to be a 1-tuple.
    return f'({",".join([f"{obj_name}.{f.name}" for f in fields])},)'


def _create_fn(name, args, body, *, globals=None, locals=None,
               return_type=MISSING):
    # Note that we mutate locals when exec() is called. Caller beware!
    if locals is None:
        locals = {}
    return_annotation = ''
    if return_type is not MISSING:
        locals['_return_type'] = return_type
        return_annotation = '->_return_type'
    args = ','.join(args)
    body = '\n'.join(f' {b}' for b in body)

    txt = f'def {name}({args}){return_annotation}:\n{body}'

    exec(txt, globals, locals)
    return locals[name]


def _field_assign(frozen, name, value, self_name):
    # If we're a frozen class, then assign to our fields in __init__
    #  via object.__setattr__.  Otherwise, just use a simple
    #  assignment.
    # self_name is what "self" is called in this function: don't
    #  hard-code "self", since that might be a field name.
    if frozen:
        return f'object.__setattr__({self_name},{name!r},{value})'
    return f'{self_name}.{name}={value}'


def _field_init(f, frozen, globals, self_name):
    # Return the text of the line in the body of __init__ that will
    #  initialize this field.

    default_name = f'_dflt_{f.name}'
    if f.default_factory is not MISSING:
        if f.init:
            # This field has a default factory.  If a parameter is
            #  given, use it.  If not, call the factory.
            globals[default_name] = f.default_factory
            value = (f'{default_name}() '
                     f'if {f.name} is _HAS_DEFAULT_FACTORY '
                     f'else {f.name}')
        else:
            # This is a field that's not in the __init__ params, but
            #  has a default factory function.  It needs to be
            #  initialized here by calling the factory function,
            #  because there's no other way to initialize it.

            # For a field initialized with a default=defaultvalue, the
            #  class dict just has the default value
            #  (cls.fieldname=defaultvalue). But that won't work for a
            #  default factory, the factory must be called in __init__
            #  and we must assign that to self.fieldname. We can't
            #  fall back to the class dict's value, both because it's
            #  not set, and because it might be different per-class
            #  (which, after all, is why we have a factory function!).

            globals[default_name] = f.default_factory
            value = f'{default_name}()'
    else:
        # No default factory.
        if f.init:
            if f.default is MISSING:
                # There's no default, just do an assignment.
                value = f.name
            elif f.default is not MISSING:
                globals[default_name] = f.default
                value = f.name
        else:
            # This field does not need initialization. Signify that to
            #  the caller by returning None.
            return None

    # Only test this now, so that we can create variables for the
    #  default.  However, return None to signify that we're not going
    #  to actually do the assignment statement for InitVars.
    if f._field_type == _FIELD_INITVAR:
        return None

    # Now, actually generate the field assignment.
    return _field_assign(frozen, f.name, value, self_name)


def _init_param(f):
    # Return the __init__ parameter string for this field.
    #  For example, the equivalent of 'x:int=3' (except instead of 'int',
    #  reference a variable set to int, and instead of '3', reference a
    #  variable set to 3).
    if f.default is MISSING and f.default_factory is MISSING:
        # There's no default, and no default_factory, just
        #  output the variable name and type.
        default = ''
    elif f.default is not MISSING:
        # There's a default, this will be the name that's used to look it up.
        default = f'=_dflt_{f.name}'
    elif f.default_factory is not MISSING:
        # There's a factory function. Set a marker.
        default = '=_HAS_DEFAULT_FACTORY'
    return f'{f.name}:_type_{f.name}{default}'


def _init_fn(fields, frozen, has_post_init, self_name):
    # fields contains both real fields and InitVar pseudo-fields.

    # Make sure we don't have fields without defaults following fields
    #  with defaults.  This actually would be caught when exec-ing the
    #  function source code, but catching it here gives a better error
    #  message, and future-proofs us in case we build up the function
    #  using ast.
    seen_default = False
    for f in fields:
        # Only consider fields in the __init__ call.
        if f.init:
            if not (f.default is MISSING and f.default_factory is MISSING):
                seen_default = True
            elif seen_default:
                raise TypeError(f'non-default argument {f.name!r} '
                                'follows default argument')

    globals = {'MISSING': MISSING,
               '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY}

    body_lines = []
    for f in fields:
        # Do not initialize the pseudo-fields, only the real ones.
        line = _field_init(f, frozen, globals, self_name)
        if line is not None:
            # line is None means that this field doesn't require
            #  initialization. Just skip it.
            body_lines.append(line)

    # Does this class have a post-init function?
    if has_post_init:
        params_str = ','.join(f.name for f in fields
                              if f._field_type is _FIELD_INITVAR)
        body_lines += [f'{self_name}.{_POST_INIT_NAME}({params_str})']

    # If no body lines, use 'pass'.
    if not body_lines:
        body_lines = ['pass']

    locals = {f'_type_{f.name}': f.type for f in fields}
    return _create_fn('__init__',
                      [self_name] +[_init_param(f) for f in fields if f.init],
                      body_lines,
                      locals=locals,
                      globals=globals,
                      return_type=None)


def _repr_fn(fields):
    return _create_fn('__repr__',
                      ['self'],
                      ['return self.__class__.__qualname__ + f"(' +
                       ', '.join([f"{f.name}={{self.{f.name}!r}}"
                                  for f in fields]) +
                       ')"'])


def _frozen_setattr(self, name, value):
    raise FrozenInstanceError(f'cannot assign to field {name!r}')


def _frozen_delattr(self, name):
    raise FrozenInstanceError(f'cannot delete field {name!r}')


def _cmp_fn(name, op, self_tuple, other_tuple):
    # Create a comparison function.  If the fields in the object are
    #  named 'x' and 'y', then self_tuple is the string
    #  '(self.x,self.y)' and other_tuple is the string
    #  '(other.x,other.y)'.

    return _create_fn(name,
                      ['self', 'other'],
                      [ 'if other.__class__ is self.__class__:',
                       f' return {self_tuple}{op}{other_tuple}',
                        'return NotImplemented'])


def _hash_fn(fields):
    self_tuple = _tuple_str('self', fields)
    return _create_fn('__hash__',
                      ['self'],
                      [f'return hash({self_tuple})'])


def _get_field(cls, a_name, a_type):
    # Return a Field object, for this field name and type.  ClassVars
    #  and InitVars are also returned, but marked as such (see
    #  f._field_type).

    # If the default value isn't derived from field, then it's
    #  only a normal default value.  Convert it to a Field().
    default = getattr(cls, a_name, MISSING)
    if isinstance(default, Field):
        f = default
    else:
        f = field(default=default)

    # Assume it's a normal field until proven otherwise.
    f._field_type = _FIELD

    # Only at this point do we know the name and the type. Set them.
    f.name = a_name
    f.type = a_type

    # If typing has not been imported, then it's impossible for
    #  any annotation to be a ClassVar. So, only look for ClassVar
    #  if typing has been imported.
    typing = sys.modules.get('typing')
    if typing is not None:
        # This test uses a typing internal class, but it's the best
        #  way to test if this is a ClassVar.
        if type(a_type) is typing.ClassVar:
            # This field is a ClassVar, so it's not a field.
            f._field_type = _FIELD_CLASSVAR

    if f._field_type is _FIELD:
        # Check if this is an InitVar.
        if a_type is InitVar:
            # InitVars are not fields, either.
            f._field_type = _FIELD_INITVAR

    # Validations for fields.  This is delayed until now, instead of
    # in the Field() constructor, since only here do we know the field
    # name, which allows better error reporting.

    # Special restrictions for ClassVar and InitVar.
    if f._field_type in (_FIELD_CLASSVAR, _FIELD_INITVAR):
        if f.default_factory is not MISSING:
            raise TypeError(f'field {f.name} cannot have a '
                            'default factory')
        # Should I check for other field settings? default_factory
        #  seems the most serious to check for. Maybe add others.  For
        #  example, how about init=False (or really,
        #  init=<not-the-default-init-value>)? It makes no sense for
        #  ClassVar and InitVar to specify init=<anything>.

    # For real fields, disallow mutable defaults for known types.
    if f._field_type is _FIELD and isinstance(f.default, (list, dict, set)):
        raise ValueError(f'mutable default {type(f.default)} for field '
                         f'{f.name} is not allowed: use default_factory')

    return f


def _find_fields(cls):
    # Return a list of Field objects, in order, for this class (and no
    #  base classes).  Fields are found from __annotations__ (which is
    #  guaranteed to be ordered).  Default values are from class
    #  attributes, if a field has a default.  If the default value is
    #  a Field(), then it contains additional info beyond (and
    #  possibly including) the actual default value.  Pseudo-fields
    #  ClassVars and InitVars are included, despite the fact that
    #  they're not real fields.  That's dealt with later.

    annotations = getattr(cls, '__annotations__', {})
    return [_get_field(cls, a_name, a_type)
            for a_name, a_type in annotations.items()]


def _set_new_attribute(cls, name, value):
    # Never overwrites an existing attribute.  Returns True if the
    #  attribute already exists.
    if name in cls.__dict__:
        return True
    setattr(cls, name, value)
    return False


def _process_class(cls, repr, eq, order, hash, init, frozen):
    # Now that dicts retain insertion order, there's no reason to use
    #  an ordered dict.  I am leveraging that ordering here, because
    #  derived class fields overwrite base class fields, but the order
    #  is defined by the base class, which is found first.
    fields = {}

    # Find our base classes in reverse MRO order, and exclude
    #  ourselves.  In reversed order so that more derived classes
    #  override earlier field definitions in base classes.
    for b in cls.__mro__[-1:0:-1]:
        # Only process classes that have been processed by our
        #  decorator.  That is, they have a _MARKER attribute.
        base_fields = getattr(b, _MARKER, None)
        if base_fields:
            for f in base_fields.values():
                fields[f.name] = f

    # Now find fields in our class.  While doing so, validate some
    #  things, and set the default values (as class attributes)
    #  where we can.
    for f in _find_fields(cls):
        fields[f.name] = f

        # If the class attribute (which is the default value for
        #  this field) exists and is of type 'Field', replace it
        #  with the real default.  This is so that normal class
        #  introspection sees a real default value, not a Field.
        if isinstance(getattr(cls, f.name, None), Field):
            if f.default is MISSING:
                # If there's no default, delete the class attribute.
                #  This happens if we specify field(repr=False), for
                #  example (that is, we specified a field object, but
                #  no default value).  Also if we're using a default
                #  factory.  The class attribute should not be set at
                #  all in the post-processed class.
                delattr(cls, f.name)
            else:
                setattr(cls, f.name, f.default)

    # Remember all of the fields on our class (including bases).  This
    #  marks this class as being a dataclass.
    setattr(cls, _MARKER, fields)

    # We also need to check if a parent class is frozen: frozen has to
    #  be inherited down.
    is_frozen = frozen or cls.__setattr__ is _frozen_setattr

    # Was this class defined with an __eq__?  Used in __hash__ logic.
    auto_hash_test= '__eq__' in cls.__dict__ and getattr(cls.__dict__, '__hash__', MISSING) is None

    # If we're generating ordering methods, we must be generating
    #  the eq methods.
    if order and not eq:
        raise ValueError('eq must be true if order is true')

    if init:
        # Does this class have a post-init function?
        has_post_init = hasattr(cls, _POST_INIT_NAME)

        # Include InitVars and regular fields (so, not ClassVars).
        flds = [f for f in fields.values()
                if f._field_type in (_FIELD, _FIELD_INITVAR)]
        _set_new_attribute(cls, '__init__',
                           _init_fn(flds,
                                    is_frozen,
                                    has_post_init,
                                    # The name to use for the "self" param
                                    #  in __init__.  Use "self" if possible.
                                    '__dataclass_self__' if 'self' in fields
                                            else 'self',
                          ))

    # Get the fields as a list, and include only real fields.  This is
    #  used in all of the following methods.
    field_list = [f for f in fields.values() if f._field_type is _FIELD]

    if repr:
        flds = [f for f in field_list if f.repr]
        _set_new_attribute(cls, '__repr__', _repr_fn(flds))

    if eq:
        # Create _eq__ method.  There's no need for a __ne__ method,
        #  since python will call __eq__ and negate it.
        flds = [f for f in field_list if f.compare]
        self_tuple = _tuple_str('self', flds)
        other_tuple = _tuple_str('other', flds)
        _set_new_attribute(cls, '__eq__',
                           _cmp_fn('__eq__', '==',
                                   self_tuple, other_tuple))

    if order:
        # Create and set the ordering methods.
        flds = [f for f in field_list if f.compare]
        self_tuple = _tuple_str('self', flds)
        other_tuple = _tuple_str('other', flds)
        for name, op in [('__lt__', '<'),
                         ('__le__', '<='),
                         ('__gt__', '>'),
                         ('__ge__', '>='),
                         ]:
            if _set_new_attribute(cls, name,
                                  _cmp_fn(name, op, self_tuple, other_tuple)):
                raise TypeError(f'Cannot overwrite attribute {name} '
                                f'in {cls.__name__}. Consider using '
                                'functools.total_ordering')

    if is_frozen:
        for name, fn in [('__setattr__', _frozen_setattr),
                         ('__delattr__', _frozen_delattr)]:
            if _set_new_attribute(cls, name, fn):
                raise TypeError(f'Cannot overwrite attribute {name} '
                                f'in {cls.__name__}')

    # Decide if/how we're going to create a hash function.
    # TODO: Move this table to module scope, so it's not recreated
    #  all the time.
    generate_hash = {(None,  False, False): ('',     ''),
                     (None,  False, True):  ('',     ''),
                     (None,  True,  False): ('none', ''),
                     (None,  True,  True):  ('fn',   'fn-x'),
                     (False, False, False): ('',     ''),
                     (False, False, True):  ('',     ''),
                     (False, True,  False): ('',     ''),
                     (False, True,  True):  ('',     ''),
                     (True,  False, False): ('fn',   'fn-x'),
                     (True,  False, True):  ('fn',   'fn-x'),
                     (True,  True,  False): ('fn',   'fn-x'),
                     (True,  True,  True):  ('fn',   'fn-x'),
                     }[None if hash is None else bool(hash),   # Force bool() if not None.
                       bool(eq),
                       bool(frozen)]['__hash__' in cls.__dict__]
    # No need to call _set_new_attribute here, since we already know if
    #  we're overwriting a __hash__ or not.
    if generate_hash == '':
        # Do nothing.
        pass
    elif generate_hash == 'none':
        cls.__hash__ = None
    elif generate_hash in ('fn', 'fn-x'):
        if generate_hash == 'fn' or auto_hash_test:
            flds = [f for f in field_list
                    if (f.compare if f.hash is None else f.hash)]
            cls.__hash__ = _hash_fn(flds)
    else:
        assert False, f"can't get here: {generate_hash}"

    if not getattr(cls, '__doc__'):
        # Create a class doc-string.
        cls.__doc__ = (cls.__name__ +
                       str(inspect.signature(cls)).replace(' -> None', ''))

    return cls


# _cls should never be specified by keyword, so start it with an
#  underscore. The presence of _cls is used to detect if this
#  decorator is being called with parameters or not.
def dataclass(_cls=None, *, init=True, repr=True, eq=True, order=False,
              hash=None, frozen=False):
    """Returns the same class as was passed in, with dunder methods
    added based on the fields defined in the class.

    Examines PEP 526 __annotations__ to determine fields.

    If init is true, an __init__() method is added to the class. If
    repr is true, a __repr__() method is added. If order is true, rich
    comparison dunder methods are added. If hash is true, a __hash__()
    method function is added. If frozen is true, fields may not be
    assigned to after instance creation.
    """

    def wrap(cls):
        return _process_class(cls, repr, eq, order, hash, init, frozen)

    # See if we're being called as @dataclass or @dataclass().
    if _cls is None:
        # We're called with parens.
        return wrap

    # We're called as @dataclass without parens.
    return wrap(_cls)


def fields(class_or_instance):
    """Return a tuple describing the fields of this dataclass.

    Accepts a dataclass or an instance of one. Tuple elements are of
    type Field.
    """

    # Might it be worth caching this, per class?
    try:
        fields =  getattr(class_or_instance, _MARKER)
    except AttributeError:
        raise TypeError('must be called with a dataclass type or instance')

    # Exclude pseudo-fields.  Note that fields is sorted by insertion
    #  order, so the order of the tuple is as the fields were defined.
    return tuple(f for f in fields.values() if f._field_type is _FIELD)


def _is_dataclass_instance(obj):
    """Returns True if obj is an instance of a dataclass."""
    return not isinstance(obj, type) and hasattr(obj, _MARKER)


def is_dataclass(obj):
    """Returns True if obj is a dataclass or an instance of a
    dataclass."""
    return hasattr(obj, _MARKER)


def asdict(obj, *, dict_factory=dict):
    """Return the fields of a dataclass instance as a new dictionary mapping
    field names to field values.

    Example usage:

      @dataclass
      class C:
          x: int
          y: int

      c = C(1, 2)
      assert asdict(c) == {'x': 1, 'y': 2}

    If given, 'dict_factory' will be used instead of built-in dict.
    The function applies recursively to field values that are
    dataclass instances. This will also look into built-in containers:
    tuples, lists, and dicts.
    """
    if not _is_dataclass_instance(obj):
        raise TypeError("asdict() should be called on dataclass instances")
    return _asdict_inner(obj, dict_factory)

def _asdict_inner(obj, dict_factory):
    if _is_dataclass_instance(obj):
        result = []
        for f in fields(obj):
            value = _asdict_inner(getattr(obj, f.name), dict_factory)
            result.append((f.name, value))
        return dict_factory(result)
    elif isinstance(obj, (list, tuple)):
        return type(obj)(_asdict_inner(v, dict_factory) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)((_asdict_inner(k, dict_factory), _asdict_inner(v, dict_factory))
                          for k, v in obj.items())
    else:
        return deepcopy(obj)


def astuple(obj, *, tuple_factory=tuple):
    """Return the fields of a dataclass instance as a new tuple of field values.

    Example usage::

      @dataclass
      class C:
          x: int
          y: int

    c = C(1, 2)
    assert astuple(c) == (1, 2)

    If given, 'tuple_factory' will be used instead of built-in tuple.
    The function applies recursively to field values that are
    dataclass instances. This will also look into built-in containers:
    tuples, lists, and dicts.
    """

    if not _is_dataclass_instance(obj):
        raise TypeError("astuple() should be called on dataclass instances")
    return _astuple_inner(obj, tuple_factory)

def _astuple_inner(obj, tuple_factory):
    if _is_dataclass_instance(obj):
        result = []
        for f in fields(obj):
            value = _astuple_inner(getattr(obj, f.name), tuple_factory)
            result.append(value)
        return tuple_factory(result)
    elif isinstance(obj, (list, tuple)):
        return type(obj)(_astuple_inner(v, tuple_factory) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)((_astuple_inner(k, tuple_factory), _astuple_inner(v, tuple_factory))
                          for k, v in obj.items())
    else:
        return deepcopy(obj)


def make_dataclass(cls_name, fields, *, bases=(), namespace=None, init=True,
                   repr=True, eq=True, order=False, hash=None, frozen=False):
    """Return a new dynamically created dataclass.

    The dataclass name will be 'cls_name'.  'fields' is an iterable
    of either (name), (name, type) or (name, type, Field) objects. If type is
    omitted, use the string 'typing.Any'.  Field objects are created by
    the equivalent of calling 'field(name, type [, Field-info])'.

      C = make_dataclass('C', ['x', ('y', int), ('z', int, field(init=False))], bases=(Base,))

    is equivalent to:

      @dataclass
      class C(Base):
          x: 'typing.Any'
          y: int
          z: int = field(init=False)

    For the bases and namespace parameters, see the builtin type() function.

    The parameters init, repr, eq, order, hash, and frozen are passed to
    dataclass().
    """

    if namespace is None:
        namespace = {}
    else:
        # Copy namespace since we're going to mutate it.
        namespace = namespace.copy()

    anns = {}
    for item in fields:
        if isinstance(item, str):
            name = item
            tp = 'typing.Any'
        elif len(item) == 2:
            name, tp, = item
        elif len(item) == 3:
            name, tp, spec = item
            namespace[name] = spec
        anns[name] = tp

    namespace['__annotations__'] = anns
    cls = type(cls_name, bases, namespace)
    return dataclass(cls, init=init, repr=repr, eq=eq, order=order,
                     hash=hash, frozen=frozen)

def replace(obj, **changes):
    """Return a new object replacing specified fields with new values.

    This is especially useful for frozen classes.  Example usage:

      @dataclass(frozen=True)
      class C:
          x: int
          y: int

      c = C(1, 2)
      c1 = replace(c, x=3)
      assert c1.x == 3 and c1.y == 2
      """

    # We're going to mutate 'changes', but that's okay because it's a new
    #  dict, even if called with 'replace(obj, **my_changes)'.

    if not _is_dataclass_instance(obj):
        raise TypeError("replace() should be called on dataclass instances")

    # It's an error to have init=False fields in 'changes'.
    # If a field is not in 'changes', read its value from the provided obj.

    for f in getattr(obj, _MARKER).values():
        if not f.init:
            # Error if this field is specified in changes.
            if f.name in changes:
                raise ValueError(f'field {f.name} is declared with '
                                 'init=False, it cannot be specified with '
                                 'replace()')
            continue

        if f.name not in changes:
            changes[f.name] = getattr(obj, f.name)

    # Create the new object, which calls __init__() and __post_init__
    #  (if defined), using all of the init fields we've added and/or
    #  left in 'changes'.
    # If there are values supplied in changes that aren't fields, this
    #  will correctly raise a TypeError.
    return obj.__class__(**changes)



def test_symbols():
    from dolang.symbols import Symbol
    import yaml

    class bcolors:
        HEADER = '\033[95m'
        OKBLUE = '\033[94m'
        OKGREEN = '\033[92m'
        WARNING = '\033[93m'
        FAIL = '\033[91m'
        ENDC = '\033[0m'
        BOLD = '\033[1m'
        UNDERLINE = '\033[4m'


    txt = yaml.safe_load(open("tests/symbols.yaml"))

    summary = []
    for short_sym, long_sym in txt['symbols'].items():


        sym = eval(long_sym)
        result = (short_sym == (str(sym)))
        summary.append([sym, long_sym, result])

        # try the other direction (match short symbol)

    for l in summary:
        ok = l[2]
        msg = bcolors.OKBLUE+'✓'+bcolors.ENDC if ok else bcolors.FAIL+'✗'+bcolors.ENDC
        print('{} | {:<50} -> {}'.format(msg, str(l[1]), l[0]))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dolang', 'dolang.tests']

package_data = \
{'': ['*']}

install_requires = \
['lark-parser>=0.9.0,<0.10.0',
 'numba>=0.50',
 'numpy>=1.18.3,<2.0.0',
 'sympy>=1.5.1,<2.0.0']

setup_kwargs = {
    'name': 'dolang',
    'version': '0.0.9',
    'description': 'Dolo Modeling Language',
    'long_description': None,
    'author': 'Winant Pablo',
    'author_email': 'pablo.winant@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import List, Dict, Tuple
import ast
from ast import NodeTransformer, Call

try:
    from .dataclasses import dataclass
except:
    from dataclasses import dataclass


from dolang.codegen import to_source
import sympy as symlib
import yaml
import io


@dataclass
class FlatFunctionFactory:
    preamble: Dict[str, str]
    content: Dict[str, str]
    arguments: Dict[str, List[str]]
    funname: str

    def __str__(self):

        d = dict(
            type="FlatFunctionFactory",
            funname=self.funname,
            preamble=self.preamble,
            content=self.content,
            arguments=self.arguments,
        )

        stream = io.StringIO()
        yaml.dump(d, stream=stream, default_flow_style=False)
        txt = stream.getvalue()
        return txt


def stack_arguments(ff: FlatFunctionFactory, varname: str = 'v'):
    import operator
    import functools
    args_k = [*ff.arguments.keys()]
    args_v = [*ff.arguments.values()]
    arguments = dict()
    arguments[varname] = functools.reduce(operator.concat, args_v[:-1])
    arguments[args_k[-1]] = args_v[-1]
    fff = FlatFunctionFactory(ff.preamble, ff.content, arguments, ff.funname)
    return fff


def substitute_preamble(ff: FlatFunctionFactory):
    import dolang
    import copy
    from .symbolic import NameSubstituter
    pr = copy.copy(ff.preamble)
    for k in pr.keys():
        pr[k] = dolang.parse_string(pr[k])
    st = NameSubstituter(pr)
    dd = copy.copy(ff.content)
    for k in dd.keys():
        dd[k] = to_source(st.transform(eq))
    return FlatFunctionFactory({}, dd, ff.arguments, ff.funname)


def get_symbolic_derivatives(fff: FlatFunctionFactory, max_order=1):

    eqs = [symlib.sympify(eq) for eq in fff.content.values()]
    varname = [*fff.arguments.keys()][0]

    svars = [symlib.sympify(v) for v in fff.arguments[varname]]

    derivatives_sym = dict()
    derivatives_sym[0] = dict(((i, ), symlib.sympify(eq))
                              for i, eq in enumerate(fff.content.values()))

    incidences = dict()
    incidence = dict()
    for i, eq in enumerate(eqs):
        ats = eq.atoms()
        l = []
        for (j, at) in enumerate(svars):
            if at in ats:
                l.append((j, at))
        incidence[(i, )] = l

    incidences[0] = incidence

    for order in range(1, max_order + 1):
        deriv = dict()
        incs = dict()
        deriv__ = derivatives_sym[order - 1]
        incs__ = incidences[order - 1]
        for eq_d, eq in deriv__.items():
            syms = incs__[eq_d]
            n = eq_d[0]
            v = eq_d[1:]
            if len(v) == 0:
                m = -1
            else:
                m = v[-1]  # max index
            for k, s in syms:
                if k >= m:
                    deq = eq.diff(s)
                    ind = eq_d + (k, )
                    deriv[ind] = deq
                    # ats = deq.atoms()
                    incs[ind] = [e for e in syms if e[1] in deq.atoms()]
        derivatives_sym[order] = deriv
        incidences[order] = incs

    return derivatives_sym
    # return derivatives, incidences



import numpy as np

from dolo.numeric.discretization import tensor_markov

class NumericEval:

    def __init__(self, d, mini_language):

        self.d = d # dictionary of substitutions
        for k,v in d.items():
            assert(isinstance(k, str))

        self.__supported_functions___ = supported_functions
        self.__supported_functions_names___ = [fun.__name__ for fun in self.__supported_functions___]

    def __call__(self, s):

        return self.eval(s)

    def eval(self, struct):

        t = struct.__class__.__name__
        method_name = 'eval_' + t.lower()
        try:
            fun = getattr(self, method_name)
        except Exception as e:
            raise Exception("Unknown type {}".format(method_name))
        return fun(struct)

    def eval_float(self, s):

        return s

    def eval_int(self, s):

        return s

    def eval_str(self, s):

        # not safe
        return eval(s, self.d)

    def eval_list(self, l):

        return [self.eval(e) for e in l]

    def eval_dict(self, d):

        if len(d) == 1:
            k = list(d.keys())[0]
            if k in self.__supported_functions_names___:
                i = self.__supported_functions_names___.index(k)
                fun = self.__supported_functions___[i]

                args = d[k]
                if isinstance(args, dict):
                    eargs = self.eval(args)
                    res = fun(**eargs)
                elif isinstance(args, (list,tuple)):
                    eargs = self.eval(args)
                    res = fun(*eargs)
                else:
                    res = args
                return res


        return {k: self.eval(e) for k,e in d.items()}

    def eval_ordereddict(self, s):

        from collections import OrderedDict
        res = OrderedDict()
        for k in s.keys():
            v = self.eval(s[k])
            res[k] = v

        return res

    def eval_ndarray(self, array_in):
        import numpy
        array_out = numpy.zeros_like(array_in, dtype=float)
        for i in range(array_in.shape[0]):
            for j in range(array_in.shape[1]):
                array_out[i,j] = self.eval(array_in[i,j])
        return array_out

    def eval_nonetype(self, none):
        return None

###
### Mini language example
###

TensorMarkov = tensor_markov

def Normal(a):
    return a

def Approximation(**kwargs):
    return {'approximation_space': kwargs}

def rouwenhorst(rho=None, sigma=None, N=None):
    from dolo.numeric.discretization import rouwenhorst
    return rouwenhorst(rho,sigma,N)

def AR1(rho, sigma, *pargs, **kwargs):
    rho_array = np.array(rho, dtype=float)
    sigma_array = np.atleast_2d( np.array(sigma, dtype=float) )
    try:
        assert(rho_array.ndim<=1)
    except:
        raise Exception("When discretizing a Vector AR1 process, the autocorrelation coefficient must be as scalar. Found: {}".format(rho_array))
    try:
        assert(sigma_array.shape[0] == sigma_array.shape[1])
    except:
        raise Exception("The covariance matrix for a Vector AR1 process must be square. Found: {}".format())
    from dolo.numeric.discretization import multidimensional_discretization
    [P,Q] = multidimensional_discretization(rho_array, sigma_array, *pargs, **kwargs)
    return P,Q

def MarkovChain(a,b):
    return [a,b]

supported_functions = [AR1, TensorMarkov, MarkovChain, Normal, Approximation]


###
###



# Markov mini language

if __name__ == '__main__':

    import numpy

    from collections import OrderedDict
    options = OrderedDict(
        smin= ['x',0.0],
        smax= ['y','x'],
        orders= [40,40],
        markov=dict(a=12.0, b=0.9)
    )


    d = {'x': 0.01, 'y': 10.0}
    print( NumericEval(d, supported_functions)(options) )



        # define a markov chain in yaml
    txt = '''
tensor:

    - rouwenhorst:
        rho: 0.9
        sigma: 0.4
        N: 3

    - markov:
        a: 0.8
        b: 1.2




    '''
    import yaml
    s = yaml.safe_load(txt)

    print(NumericEval(d, supported_functions)(s))



import numpy
import ast


def eval_with_diff(f, args, add_args, epsilon=1e-8):

    # f is a guvectorized function: f(x1, x2, ,xn, y1,..yp)
    # args is a list of vectors [x1,...,xn]
    # add_args is a list of vectors [y1,...,yn]
    # the function returns a list [r, dx1, ..., dxn] where:
    # r is the vector value value of f at (x1, xn, y1, yp)
    # dxi is jacobian w.r.t. xi

    # TODO: generalize when x1, ..., xn have non-core dimensions

    epsilon = 1e-8
    vec = numpy.concatenate(args)
    N = len(vec)
    points = vec[None, :].repeat(N+1, axis=0)
    for i in range(N):
        points[1+i, i] += epsilon

    argdims = [len(e) for e in args]
    cn = numpy.cumsum(argdims)
    slices = [e for e in zip([0] + cn[:-1].tolist(), cn.tolist())]
    vec_args = tuple([points[:, slice(*sl)] for sl in slices])

    arg_list = vec_args + add_args
    jac = f(*arg_list)
    res = jac[0, :]
    jac[1:, :] -= res[None, :]
    jac[1:, :] /= epsilon
    jacs = [jac[slice(sl[0]+1, sl[1]+1), :] for sl in slices]
    jacs = [j.T.copy() for j in jacs]  # to get C order
    return [res] + jacs


class standard_function:

    epsilon = 1e-8

    def __init__(self, fun, n_output):

        # fun is a vectorized, non-allocating function
        self.fun = fun
        self.n_output = n_output

    def __call__(self, *args, diff=False, out=None):

        non_core_dims = [ a.shape[:-1] for a in args]
        core_dims = [a.shape[-1:] for a in args]

        non_core_ndims = [len(e) for e in non_core_dims]

        if (max(non_core_ndims) == 0):
            # we have only vectors, deal wwith it directly
            if not diff:
                if out is None:
                     out = numpy.zeros(self.n_output)
                self.fun(*(args+(out,)))
                return out

            else:
                def ff(*aa):
                    return self.__call__(*aa, diff=False)
                n_ignore = 1 # number of arguments that we don't differentiate
                res = eval_with_diff(ff, args[:-n_ignore], args[-n_ignore:], epsilon=1e-8)
                return res


        else:

            if not diff:
                K = max( non_core_ndims )
                ind = non_core_ndims.index( K )
                biggest_non_core_dim = non_core_dims[ind]
                biggest_non_core_dims = non_core_ndims[ind]
                new_args = []
                for i,arg in enumerate(args):
                    coredim = non_core_dims[i]
                    n_None = K-len(coredim)
                    n_Ellipsis = arg.ndim
                    newind = ((None,)*n_None) +(slice(None,None,None),)*n_Ellipsis
                    new_args.append(arg[newind])

                new_args = tuple(new_args)
                if out is None:
                    out = numpy.zeros( biggest_non_core_dim + (self.n_output,) )

                self.fun(*(new_args + (out,)))
                return out

            else:
                # older implementation
                return self.__vecdiff__(*args, diff=True, out=out)

    def __vecdiff__(self,*args, diff=False, out=None):


        fun = self.fun
        epsilon = self.epsilon

        sizes = [e.shape[0] for e in args if e.ndim==2]
        assert(len(set(sizes))==1)
        N = sizes[0]

        if out is None:
            out = numpy.zeros((N,self.n_output))

        fun( *( list(args) + [out] ) )

        if not diff:
            return out
        else:
            l_dout = []
            for i, a in enumerate(args[:-1]):
                # TODO: by default, we don't diffferentiate w.r.t. the last
                # argument. Reconsider.
                pargs = list(args)
                dout = numpy.zeros((N, self.n_output, a.shape[1]))
                for j in range( a.shape[1] ):
                    xx = a.copy()
                    xx[:,j] += epsilon
                    pargs[i] = xx
                    fun(*( list(pargs) + [dout[:,:,j]]))
                    dout[:,:,j] -= out
                    dout[:,:,j] /= epsilon
                l_dout.append(dout)
            return [out] + l_dout



import copy
import ast
from ast import NodeVisitor

from .symbolic import eval_scalar


def parse(s): return ast.parse(s).body[0].value

def unique(seq):
    seen = set()
    for item in seq:
        if item not in seen:
            seen.add(item)
            yield item

def tshift(t, n):
    return (t[0], t[1]+n)

def get_atoms(string):

    expr = ast.parse(str.strip(str(string)))
    parser = FindNames()
    parser.visit(expr)
    names = parser.names

    return set(names)

class FindNames(NodeVisitor):

    def __init__(self):
        self.names = []

    def visit_Name(self, node):
        self.names.append(node.id)

def triangular_solver(incidence, context=None):

    if context is None:

        context = dict()
        import math
        context['log'] = math.log
        context['exp'] = math.exp
        context['nan'] = float('nan')

    n = len(incidence)

    current = copy.deepcopy(incidence)

    solved = []
    max_steps = len(incidence)
    steps = 0

    while (steps < max_steps) and len(solved) < n:

        possibilities = [i for i in range(n) if (
            i not in solved) and (len(current[i]) == 0)]

        for i in possibilities:
            for e in current:
                if i in e:
                    e.remove(i)
            solved.append(i)

        steps += 1

    if len(solved) < n:
        raise Exception('System is not triangular')
    return solved


def get_incidence(sdict):

    var_order = list(sdict.keys())
    var_set = set(var_order)
    expressions = [sdict[k] for k in var_order]
    incidence = []

    for i, eq in enumerate(expressions):

        atoms = get_atoms(eq)
        vars = var_set.intersection(atoms)
        inds = [var_order.index(v) for v in vars]
        incidence.append(inds)

    return incidence


from collections import OrderedDict


def solve_triangular_system(system, values=None, context=None):

    system = OrderedDict(system)
    var_order = list(system.keys())
    ll = get_incidence(system)
    sol_order = triangular_solver(ll, context=context)
    d = copy.copy(values) if values else {}

    import math
    d['nan'] = float('nan')
    d['exp'] = math.exp
    d['log'] = math.log
    d['sin'] = math.sin
    d['cos'] = math.cos

    for i in sol_order:
        v = var_order[i]
        try:
            val = system[v]
            d[v] = eval(str(val), d, d)
        except Exception as e:  # in case d[v] is an int
            raise(e)

    resp = OrderedDict([(v, d[v]) for v in system.keys()])
    return resp


def get_deps(incidence, var, visited=None):

    # assert(var in incidence)
    assert(isinstance(var, tuple) and len(var) == 2)

    if visited is None:
        visited = (var,)
    elif var in visited:
        raise Exception("Non triangular system.")
    else:
        visited = visited + (var,)

    n = var[1]
    if abs(n) > 20:
        raise Exception("Found variable with time {}. Something has probably gone wrong.".format(n))

    deps = incidence[(var[0], 0)]
    if n != 0:
        deps = [tshift(e, n) for e in deps]

    resp = sum([get_deps(incidence, e, visited) for e in deps], [])

    resp.append(var)

    return resp



def test_triangular_solution():

    d = {
        'a': 0.1,
        'b': 'a + 3',
        'c': 'a + b'
    }
    sol = solve_triangular_system(d)
    for v in sol.values():
        assert( isinstance(v, float))

if __name__ == '__main__':
    test_triangular_solution()



def test_patterns():

    import dolang
    from dolang.pattern import compare_strings
    import yaml

    class bcolors:
        HEADER = '\033[95m'
        OKBLUE = '\033[94m'
        OKGREEN = '\033[92m'
        WARNING = '\033[93m'
        FAIL = '\033[91m'
        ENDC = '\033[0m'
        BOLD = '\033[1m'
        UNDERLINE = '\033[4m'

    tests = yaml.safe_load(open('tests/patterns.yaml'))

    pattern_tests = tests['patterns']

    for l in pattern_tests:

        res = compare_strings(l[0], l[1])
        expected = l[2]
        ok = (res==expected)
        msg = bcolors.OKBLUE+'✓'+bcolors.ENDC if ok else bcolors.FAIL+'✗'+bcolors.ENDC
        print('{} : {} : {} : ({} instead of {})'.format(msg,l[0],l[1],res,expected))
